// ########################################################################
// ## bipart.cc                                                          ##
// ## Zufllige Erzeugung bipartiter von Graphen                         ##
// ## Copyright Thomas Dreibholz, uzs6pq@ibm.rhrz.uni-bonn.de            ##
// ##           WWW Homepage:     http://ibm.rhrz.uni-bonn.de/~uzs6pq    ##
// ########################################################################

#define BipartName "Bipart 1.0"

#include <stdio.h>
#include <stdlib.h>
#include <iostream.h>
#include <fstream.h>
#include <string.h>

#include "../tdgf/tdgf.h"
#include "../tdgf/graph.h"
#include "../tdgf/graph.icc"
#include "../zufall/zufall.h"


// Initialisierung des Kantenarrays:
// grad[] ist ein Array der Lnge knoten mit den Gradzahlen der Knoten 
// In menge[] wird eine Liste von grad[i] Labels des i-ten Knoten fr
// i in aufsteigender Ordnung erstellt.
// Also z.B. 1,1,1,2,3,3 fr Gradzahlen 3 1 2
void Initialisiere(card32 knoten, card32 start, card32* grad, card32* menge)
{
   card32 i,j,k;

   k = 0;
   for(i = 1;i <= knoten;i++) {
      for(j = 0;j < grad[i];j++) {
         k++;
         menge[k] = i + (start - 1);
      }      
   }
}


// Erzeugung eines bipartiten Graphen:
// knotenRot und knotenBlau enthalten die Anzahlen der roten bzw. blauen Knoten
// kanten ist die Anzahl der Kanten
// gradRot[i] und gradBlau[i] enthalten die Gradzahl der i-ten blauen
//    bzw. roten Knotens
// Nach der Ausfhrung sind die Kanten des erzeugten Graphen in folgender
//    Weise gespeichert: (rot[1],blau[1]), ..., (rot[kanten],blau[kanten])
bool Bipart(card32 knotenRot, card32 knotenBlau, card32 kanten,
            card32* gradRot, card32* gradBlau, card32* rot, card32* blau,
            Graph* graph)
{
   bool   good;
   card32 i,t,x;
   card32 zaehler = 0;

   Initialisiere(knotenRot,1,gradRot,rot);
   Initialisiere(knotenBlau,knotenRot + 1,gradBlau,blau);

   good = false;

   while(!good) {
      // Zhler berprfen. Falls zu gro -> Bipartiter Graph ist mit
      // angegebenen Daten mit hoher Wahrscheinlichkeit nicht mglich
      zaehler++;
      if(zaehler >= (kanten * 10)) {
         return(false);
      }

      // Neuen Versuch starten, einen bipartiten Graphen zu erzeugen
      graph->Neu();
      good = true;
      i = kanten;
      while((i >= 1) && good) {
         // x wird zufllig aus {1,...,i} gewhlt.
         x = Random(i);
         // Auf multi-edges berprfen
         if(graph->Verbunden(rot[i] - 1,blau[x] - 1)) {
            good = false; 
         }
         else {
            t = blau[x];
            blau[x] = blau[i];
            blau[i] = t;
            graph->Verbinde(rot[i] - 1,blau[i] - 1);
            i--;     
         }
      }
   }

   // Labels und Bipart-Einstellungen sichern
   char label[16];
   for(i = 1;i <= knotenRot;i++) {
      graph->SetzeBipart(i - 1,true);
      sprintf(label,"%d",i);
      graph->SetzeLabel(i - 1,label);
   }
   for(i = 1;i <= knotenBlau;i++) {
      graph->SetzeBipart(i + knotenRot - 1,false);
      sprintf(label,"%d",i);
      graph->SetzeLabel(i + knotenRot - 1,label);
   }

   return(true);
}


// Hauptprogramm
int32 main(card32 argc, char** argv)
{
   card32 knoten        = 0;
   card32 kstart        = 0;
   card32 kende         = 0;
   card32 gstart        = 1;
   card32 gende         = 4;
   card32 anzahl        = 1;
   card32 erfolglos     = 0;
   card32 erfolg        = 0;
   card32 abbruchAnzahl = 0;
   bool bitMatrix   = true;
   char* titel      = NULL;
   char* copyright  = NULL;
   char* programm   = BipartName;
   char* name       = NULL;
   bool  ausgabe    = false;
   ofstream           os;
   card32             i,j;

   if(argc == 1) {
      cerr << "Aufruf: " << argv[0] << " {Parameter}" << endl;
      cerr << "-a[#]    - Minimale Knotenzahl setzen" << endl;
      cerr << "-b[#]    - Maximale Knotenzahl setzen" << endl;
      cerr << "-k[#]    - Knotenzahl setzen (entspricht min == max)" << endl;
      cerr << "-g[#]    - Minimale Gradzahl setzen" << endl;
      cerr << "-h[#]    - Maximale Gradzahl setzen" << endl;
      cerr << "-#[#]    - Anzahl der zu erzeugenden Graphen" << endl;
      cerr << "-s[#]    - Maximale Anzahl von Fehlversuchen" << endl;
      cerr << "-f[Name] - Dateiname fr Ausgabe (TDGF)" << endl;
      cerr << "-t[Text] - Titel setzen" << endl;
      cerr << "-c[Text] - Copyright setzen" << endl;
      cerr << "-y       - Adjazenzmatrix im Byte-Format statt Bit-Format speichern" << endl;
      cerr << "-$       - Ausgabe zustzlich auf Bildschirm" << endl;
      exit(1);
   }

   // Parameter auswerten
   for(i = 1;i < argc;i++) {
      if(!(strncmp(argv[i],"-t",2)))      titel     = &argv[i][2];
      else if(!(strncmp(argv[i],"-c",2))) copyright = &argv[i][2];
      else if(!(strncmp(argv[i],"-a",2))) kstart    = atoi(&argv[i][2]);
      else if(!(strncmp(argv[i],"-b",2))) kende     = atoi(&argv[i][2]);
      else if(!(strncmp(argv[i],"-k",2))) knoten    = atoi(&argv[i][2]);
      else if(!(strncmp(argv[i],"-g",2))) gstart    = atoi(&argv[i][2]);
      else if(!(strncmp(argv[i],"-h",2))) gende     = atoi(&argv[i][2]);
      else if(!(strncmp(argv[i],"-s",2))) abbruchAnzahl = atoi(&argv[i][2]);
      else if(!(strncmp(argv[i],"-#",2))) anzahl    = atoi(&argv[i][2]);
      else if(!(strncmp(argv[i],"-f",2))) name      = &argv[i][2];
      else if(!(strcmp(argv[i],"-y")))    bitMatrix = false;
      else if(!(strcmp(argv[i],"-$")))    ausgabe = true;  
   } 
   if(name == NULL) ausgabe = true;
   if(anzahl == 0) anzahl = 1;
   if(knoten == 0) {
      if(kstart < 1) kstart = 1;
      if(kende < kstart) kende = kstart;
   }
   else {
      if(knoten < 1) knoten = 1;
      kstart = knoten;
      kende  = knoten;
   }
   if(gstart < 1) gstart = 1;
   if(gende < gstart) gende = gstart;
   if(titel == NULL)
      titel = "Zufllig erzeugte bipartite Graphen";
   if(name) {
      if(!ErstelleTDGF(os,name,titel,copyright,programm)) {
         cerr << "Kann Datei \"" << name << "\" nicht erstellen!" << endl;
         exit(1);
      }
   }
   if(abbruchAnzahl == 0) abbruchAnzahl = 20 * anzahl;
   if(abbruchAnzahl < 5 * anzahl)  abbruchAnzahl = 5 * anzahl;


   // Zufallszahlengenerator initialisieren
   Randomize();


   // Informationen ausgeben
   cout << "Parameter fr die Generierung:" << endl;
   cout << "Knotenzahlen von " << kstart << " bis " << kende << endl;
   cout << "Gradzahlen von " << gstart << " bis " << gende << endl;
   cout << "Maximal " << abbruchAnzahl << " mgliche Fehlversuche" << endl;
   cout << "Anzahl:    " << anzahl << endl;
   cout << "Titel:     " << titel << endl;
   cout << "Copyright: " << copyright << endl << endl;
   
   j = 0;
   while(j < anzahl) {
      card32 kanten;
      card32 kantenRot;
      card32 kantenBlau;
      card32 knotenRot;
      card32 knotenBlau;
      card32* gradRot  = NULL;
      card32* gradBlau = NULL;

      do {
         knotenRot  = Random(1 + kende - kstart) + kstart - 1;
         knotenBlau = Random(1 + kende - kstart) + kstart - 1;
         if(gradRot) delete gradRot;
         if(gradBlau) delete gradBlau;
         gradRot   = new card32[knotenRot + 1];
         gradBlau  = new card32[knotenBlau + 1];

         kantenRot  = 0;
         kantenBlau = 0;
         for(i = 1;i <= knotenRot;i++) {
            gradRot[i] = Random(1 + min(gende,knotenBlau) - gstart) + gstart - 1;
            kantenRot += gradRot[i];
         }
         for(i = 1;i <= knotenBlau;i++) {
            if(kantenBlau < kantenRot) {
               gradBlau[i] = Random(min(kantenRot - kantenBlau,knotenRot));
               kantenBlau += gradBlau[i];
            }
            else { gradBlau[i] = 1; kantenBlau++; }
         }
         kanten = kantenRot;
      } while(kantenRot != kantenBlau);

      // Daten ausgeben
      if(ausgabe == true) {
         cout << "#" << j + 1 << ":" << endl;
         cout << "Knoten:    rot = " << knotenRot << "   blau = " << knotenBlau << endl;   
         cout << "Kanten:    " << kanten << endl;
         cout << "Grad rot:  ";
         for(i = 1;i <= knotenRot;i++) cout << gradRot[i] << "  ";
         cout << endl << "Grad blau: ";
         for(i = 1;i <= knotenBlau;i++) cout << gradBlau[i] << "  ";
         cout << endl;
      }

      // Graph erzeugen
      card32* rot  = new card32[kanten + 1];
      card32* blau = new card32[kanten + 1];
      for(i = 0;i <= kanten;i++)
         rot[i] = blau[i] = 0;

      Graph graph(knotenRot + knotenBlau,bitMatrix ? BIT_MATRIX : BYTE_MATRIX);
      if(Bipart(knotenRot,knotenBlau,kanten,gradRot,gradBlau,rot,blau,&graph)) {    
         if(ausgabe) cout << graph << endl;
         if(name) graph.Schreibe(os);
         erfolg++;
         j++;
      }
      else {
         if(ausgabe) {
            cout << "Mit diesen Daten ist die Erzeugung eines bipartiten Graphen mit hoher" << endl;
            cout << "Wahrscheinlichkeit nicht mglich!" << endl << endl;
         }
         erfolglos++;
         if(erfolglos > abbruchAnzahl) {
            cerr << "Abbruch! Startparameter ergaben zu viele erfolglose Versuche!" << endl;
            j = anzahl;
         }
      }    

      delete rot;
      delete blau;
      delete gradRot;
      delete gradBlau;
   }

   cout << "Erzeugte Graphen:    " << erfolg << endl;
   cout << "Erfolglose Versuche: " << erfolglos << endl;
 
   return 0;
}
